from sqlite3.dbapi2 import DatabaseError, Error
from statemachine import StateMachine, State

from pkg_trainmote.actions.actionInterface import ActionInterface
from .models.Program import Program
from .models.Action import Action
from typing import Optional, List
from .actions import actionHelper

class ProgramMachine(StateMachine):

    program: Optional[Program] = None
    __action_index: int = 0
    __actionInterface: Optional[ActionInterface] = None
    __currentAction: Optional[Action] = None

    isRunning: bool = False

    readyForAction = State('readyForAction', initial=True)
    runningAction = State('RunningAction')
    actionFinished = State('ActionFinished')
    preparingAction = State('preparingAction')

    startProgram = readyForAction.to(preparingAction)
    cancelProgram = runningAction.to(readyForAction)
    endAction = runningAction.to(actionFinished)
    prepareForAction = actionFinished.to(preparingAction)
    startAction = preparingAction.to(runningAction)
    endProgram = preparingAction.to(readyForAction)

    def start(self, program: Program):
        if program.name is not None:
            print("Start " + program.name)

        if self.is_readyForAction:
            self.program = program
            self.startProgram()
        else:
            raise Error("A program is already running.")

    def on_startProgram(self):
        self.isRunning = True
        self.__currentAction = self.loadAction()

    def on_enter_runningAction(self):
        def actionCallback():
            self.endAction()

        self.__actionInterface.runAction(actionCallback)

    def on_cancelProgram(self):
        self.isRunning = False
        self.program = None
        if self.__actionInterface is not None:
            self.__actionInterface.cancelAction()
        self.__action_index = 0

    def on_endAction(self):
        self.__action_index = self.__action_index + 1
        self.__currentAction = self.loadAction()

    def on_enter_actionFinished(self):
        self.prepareForAction()

    def on_enter_preparingAction(self):
        self.prepare()

    def prepare(self):
        if self.__currentAction is not None:
            self.__actionInterface = actionHelper.getProgramAction(self.__currentAction)
            if self.__actionInterface is not None:
                self.__actionInterface.prepareAction()
                self.startAction()
                return
        self.endProgram()

    def on_endProgram(self):
        print('Program End')
        self.isRunning = False
        self.program = None
        self.__action_index = 0

##
#
##
    def getCurrentAction(self) -> Optional[Action]:
        return self.__currentAction

##
#   Load the action for the current __action_index.
##
    def loadAction(self) -> Optional[Action]:
        if self.program is not None and self.__action_index < len(self.program.actions):
            return self.program.actions[self.__action_index]
        return None

##
# Returns the following actions of the program
##
    def followingActions(self) -> List[Action]:
        if self.program is not None and self.__action_index < len(self.program.actions):
            return self.program.actions[self.__action_index + 1:]
        return []



import threading
from typing import Optional
from pkg_trainmote.models.Action import Action
from pkg_trainmote.actions.actionInterface import ActionInterface
import time
from datetime import datetime, timedelta, date

class ClockThread(threading.Thread):

    mDate = None

    def __init__(self, mDate: datetime, callback):
        threading.Thread.__init__(self)
        self.mDate = mDate
        self.kill = threading.Event()
        self.callback = callback

    def run(self):
        if self.mDate is not None:
            now = datetime.now()
            while now < self.mDate:
                diff = (self.mDate - now).total_seconds()
                if diff > 60:
                    time.sleep(60)
                elif diff > 0:
                    time.sleep(1)
                now = datetime.now()

            if self.callback is not None:
                self.callback()


class ClockAction(ActionInterface):

    __action: Optional[Action] = None
    timer: Optional[ClockThread] = None
    __date: datetime = datetime.now()

    def __init__(self, action: Action) -> None:
        self.__action = action

    def prepareAction(self):
        if self.__action is not None:
            hhValue = int(self.__action.inputs[0])
            hour = 0
            if hhValue <= 23 and hhValue >= 0:
                hour = hhValue

            mmValue = int(self.__action.inputs[1])
            minutes = 0
            if mmValue <= 59 and mmValue >= 0:
                minutes = mmValue

            ssValue = int(self.__action.inputs[1])
            seconds = 0
            if ssValue <= 59 and ssValue >= 0:
                seconds = ssValue

            now = datetime.now()
            self.__date = datetime(now.year, now.month, now.day, hour, minutes, seconds)
            if self.__date < datetime.now():
                tomorrow = date.today() + timedelta(days=1)
                self.__date = datetime(tomorrow.year, tomorrow.month, tomorrow.day, hour, minutes, seconds)

            print(self.__date)

    def runAction(self, _callback):
        self.timer = ClockThread(self.__date, _callback)
        self.timer.start()

    def cancelAction(self):
        self.timer.join()
        self.timer.kill.set()
        pass


# class BaseThread(threading.Thread):
#     def __init__(self, callback=None, callback_args=None, *args, **kwargs):
#         target = kwargs.pop('target')
#         super(BaseThread, self).__init__(target=self.target_with_callback, *args, **kwargs)
#         self.callback = callback
#         self.method = target
#         self.callback_args = callback_args

#     def target_with_callback(self):
#         self.method()
#         if self.callback is not None:
#             self.callback(*self.callback_args)



from typing import Optional, Dict, Any, List
import enum

class Action():

    def __init__(
        self,
        uid: Optional[str],
        type: Optional[str],
        position: Optional[int],
        inputs: Optional[List[str]],
        name: Optional[str]
    ):
        self.uid = uid
        self.type = type
        self.position = position
        self.name = name
        self.inputs = inputs

    def to_dict(self):
        return {
            "uid": self.uid,
            "type": self.type,
            "position": self.position,
            "inputs": self.inputs,
            "name": self.name
        }

    @classmethod
    def from_dict(cls, data: Dict[str, Any]):
        if "inputs" in data.keys():
            inputs: List[str] = []
            for value in data.get("inputs"):
                inputs.append(value)
            return cls(
                None if data.get("uid") is None else str(data.get("uid")),
                None if data.get("type") is None else str(data.get("type")),
                None if data.get("position") is None else int(str(data.get("position"))),
                inputs,
                data.get("name")
            )
        else:
            return cls(
                None if data.get("uid") is None else str(data.get("uid")),
                None if data.get("type") is None else str(data.get("type")),
                None if data.get("position") is None else int(str(data.get("position"))),
                None,
                data.get("name")
            )


class ActionType(enum.Enum):
    TM_SET_SWITCH = "SetSwitch"
    TM_SET_STOP = "SetStop"
    TM_PERFORM_TIMER = "WaitFor"
    TM_PERFORM_ALARM = "StartAtClock"



from typing import Optional
from pkg_trainmote.models.Program import Program
from .programMachine import ProgramMachine

programMachine = ProgramMachine()

##
# Method to get current running program. If no program running method returns None.
##
def getRunningProgramm() -> Optional[Program]:
    if programMachine.isRunning:
        return programMachine.program
    else:
        return None
